/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include "replayk.h"

//#define DO_DATA_COPY

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/namei.h>
#include <linux/writeback.h>
#include <linux/kthread.h>
#include "prefetch.h"

#if MULTITHREADING
extern int thread_execute(void* arg);
extern int init_thread_loc;
#endif

static inline struct dentry *lock_parent(struct dentry *dentry)
{
    struct dentry *dir = dget(dentry->d_parent);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
    down(&dir->d_inode->i_sem);
#else
    mutex_lock(&dir->d_inode->i_mutex);
#endif
    return dir;
}

static inline void unlock_dir(struct dentry *dir)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
    up(&dir->d_inode->i_sem);
#else
    mutex_unlock(&dir->d_inode->i_mutex);
#endif
    dput(dir);
}

//Returns the full pathname of the dentry - OBSOLETE
static char* path_from_dentry(struct dentry* dir)
{
  struct dentry* dptr;
  int len;
  int pos;
  char* path = NULL;

  ASSERT(dir != NULL);

  len = dir->d_name.len + 1; //+1 for root '/'
  dptr = dir->d_parent;

  //Find Length first
  while(strncmp(dptr->d_name.name,"/",dptr->d_name.len)) {
    len += dptr->d_name.len + 1; //+1 for '/'
    dptr = dptr->d_parent;
  }

  path = KMALLOC(len+1,GFP_KERNEL);
  if(path == NULL) {
    printk(KERN_ALERT "%s() - out of memory\n",__FUNCTION__);
    return NULL;
  }
  path[len] = (char)NULL;
  pos = len;
  for(dptr = dir;strncmp(dptr->d_name.name,"/",dptr->d_name.len);dptr = dptr->d_parent) {
    pos -= dptr->d_name.len;
    ASSERT(pos >= 1);
    memcpy(path+pos,dptr->d_name.name,dptr->d_name.len);
    path[--pos] = '/';
  }

  ASSERT(!pos);

  return path;
}

#endif

extern spinlock_t dcache_lock;

// this is a macro to check any return value that is a pointer from the VFS call

#define CHECK_RETURN_PTR(ptr, retval) do { \
		if ((IS_ERR((ptr)) || (retval) != 0) &&  (ptr) != ERR_PTR((retval))) { \
			printk(KERN_WARNING"%s() - Unexpected error during replay(p): %p,expected:%x\n", __FUNCTION__,(ptr),((int)retval)); \
			err = -EIO; \
			goto out; \
		} \
	} while (0);

#define CHECK_RETURN_INT(ival, retval) do { \
		if ((ival) != (retval)) { \
			printk(KERN_WARNING"%s() - Unexpected error during replay(i): %i,expected:%i\n", __FUNCTION__,(ival),(retval)); \
			err = -EIO; \
			goto out; \
		} else \
			err = 0; \
	} while (0);

#define DENTRY_ARG(arg) ratp[ srcmdp->args[(arg)] ].u.dentry
#define MNT_ARG(arg) ratp[ srcmdp->args[(arg)] ].u.mnt
#define FILE_ARG(arg) ratp[ srcmdp->args[(arg)] ].u.file
#define STR_ARG(arg)  ratp[ srcmdp->args[(arg)] ].u.cpval
#define R_INT_ARG(arg)  ratp[ srcmdp->args[(arg)] ].u.ival
#define INT_ARG(arg) srcmdp->args[arg]

// replay write command
int replay_write(struct replay_command* srcmdp, struct replay* r)
{
	int err = 0;
	struct res_alloc_table* ratp;
	struct file *f;
	mm_segment_t oldfs;

	print_entry_location();

	ratp = r->ratp;
	oldfs = get_fs();

	CHECK_RES_TYPE(srcmdp->args[0], ARG_FILE);
	CHECK_RES_TYPE(srcmdp->args[1], ARG_STR);

	f = FILE_ARG(0);

	/*printk(KERN_WARNING"before: i_blksize:%lu i_blkbits:%u i_blocks:%lu, i_bytes:%u, i_size:%lu\n",
	     f->f_dentry->d_inode->i_blksize, f->f_dentry->d_inode->i_blkbits,
	     f->f_dentry->d_inode->i_blocks, f->f_dentry->d_inode->i_bytes, (long unsigned int)f->f_dentry->d_inode->i_size);*/

	set_fs(KERNEL_DS);
	err = f->f_op->write(f, STR_ARG(1), INT_ARG(2), &(f->f_pos));
	set_fs(oldfs);

	/*printk(KERN_WARNING"after: i_blksize:%lu i_blkbits:%u i_blocks:%lu, i_bytes:%u, i_size:%lu\n",
	     f->f_dentry->d_inode->i_blksize, f->f_dentry->d_inode->i_blkbits,
	     f->f_dentry->d_inode->i_blocks, f->f_dentry->d_inode->i_bytes, (long unsigned int)f->f_dentry->d_inode->i_size);*/

	CHECK_RETURN_INT(err, srcmdp->expected_retval);

out:
	print_exit_status(err);
	return err;
}

// dummy replayfs read actor that doesn't copy any data
int replayfs_file_read_actor(read_descriptor_t *desc, struct page *page,
		unsigned long offset, unsigned long size)
{
#ifdef DO_DATA_COPY
	char *kaddr;
	unsigned long left;
#endif
	unsigned long count = desc->count;

	if (size > count)
		size = count;

#ifdef DO_DATA_COPY
	/*
	 * Faults on the destination of a read are common, so do it before
	 * taking the kmap.
	 */
	if (!fault_in_pages_writeable(desc->arg.buf, size)) {
		kaddr = kmap_atomic(page, KM_USER0);
		left = __copy_to_user_inatomic(desc->arg.buf, kaddr + offset, size);
		kunmap_atomic(kaddr, KM_USER0);
		if (left == 0)
			goto success;
	}

	/* Do it the slow way */
	kaddr = kmap(page);
	left = __copy_to_user(desc->arg.buf, kaddr + offset, size);
	kunmap(page);

	if (left) {
		size -= left;
		desc->error = -EFAULT;
	}
#endif
success:
	desc->count = count - size;
	desc->written += size;
#ifdef DO_DATA_COPY
	desc->arg.buf += size;
#endif
	return size;
}

// replay read command
int replay_read(struct replay_command* srcmdp, struct replay* r)
{
	int err = 0;
	struct file *f;
#ifdef DO_DATA_COPY
	char* buffer = NULL;
 	mm_segment_t oldfs;
#endif
	struct res_alloc_table* ratp;
	read_descriptor_t desc;
	loff_t pos = 0;

	print_entry_location();

#ifdef DO_DATA_COPY
	oldfs = get_fs();
	buffer = NULL;
#endif

	ratp = r->ratp;

	CHECK_RES_TYPE(srcmdp->args[0], ARG_FILE);
	//This is 0 if read capturing was disabled
/* 	if(srcmdp->args[1]) */
/* 	  CHECK_RES_TYPE(srcmdp->args[1], ARG_STR); */

	f = FILE_ARG(0);
/* 	pos = (loff_t*)STR_ARG(3); */
	//*pos = 0; //XXX: temp hack.

#ifdef DO_DATA_COPY
	buffer = VMALLOC(1024); // pread hack
	if (!buffer) {
		err = -ENOMEM;
		goto out;
	}

	set_fs(KERNEL_DS);
	//err = f->f_op->read(f, buffer, INT_ARG(2), &(f->f_pos));
	set_fs(oldfs);
#endif
	desc.written = 0;
	desc.arg.buf = (void*)0xACDC;
	desc.count = INT_ARG(2);
	if (desc.count != 0) {
		desc.error = 0;
		do_generic_file_read(f, &pos, &desc, replayfs_file_read_actor);
 		err = desc.written;
		if (!err)
			err = desc.error;
		CHECK_RETURN_INT(err, srcmdp->expected_retval);
	}
#ifdef DO_DATA_COPY
	set_fs(oldfs);
#endif
	/* if(srcmdp->args[1]) {
	  if (r->b_compare_data && strcmp(buffer, STR_ARG(1)) != 0) {
	    printk(KERN_WARNING"Data comparison error\n");
	    err = -EIO;
	    goto out;
	  }
	} */
out:
#ifdef DO_DATA_COPY
	if (buffer)
		VFREE(buffer);
#endif

	print_exit_status(err);
	return err;
}

// replay flush command
int replay_flush(struct replay_command* srcmdp, struct replay* r)
{
  int err = 0;
  struct file* f;
  struct res_alloc_table* ratp = r->ratp;

  print_entry_location();

  CHECK_RES_TYPE(srcmdp->args[0], ARG_FILE);

  f = FILE_ARG(0);

  if(f->f_op && f->f_op->flush) {
    lock_kernel();
    fist_print_dentry("before flush", f->f_dentry);
    err = f->f_op->flush(f);
    fist_print_dentry("after flush", f->f_dentry);
    unlock_kernel();
  }
  else
    goto out;

  CHECK_RETURN_INT(err, srcmdp->expected_retval);
 out:
  print_exit_status(err);
  return err;
}

// replay release command
int replay_release(struct replay_command* srcmdp, struct replay* r)
{
	int err = 0;
	int dr;
	struct file* f;
	struct dentry *d;
	struct res_alloc_table* ratp = r->ratp;

	print_entry_location();

	CHECK_RES_TYPE(srcmdp->args[0], ARG_FILE);

	f = FILE_ARG(0);
	dr = lookup_rat(r, (int)f->f_dentry, ARG_DENTRY);
	if (dr > 0 && ratp[dr].count > 0) {
		ratp[dr].count--;
	//	if (ratp[dr].count == 0)
	//		ratp[dr].dentry = NULL;
	}

	d = f->f_dentry;
	fist_print_dentry("before fput", d);
	fput(f);
	ratp[RAT_ROOT_MNT].count--;
	fist_print_dentry("after fput", d);

	ratp[srcmdp->args[0]].count--;
	if (ratp[srcmdp->args[0]].count == 0) {
		ratp[srcmdp->args[0]].type = ARG_INVAL;
		FILE_ARG(0) = NULL;
	}

out:
	print_exit_status(err);
	return err;
}

// replay lookup command
int replay_lookup(struct replay_command* srcmdp, struct replay* r)
{
	struct dentry *new_dentry;
	int err = 0;
	struct res_alloc_table* ratp = r->ratp;

	print_entry_location();

	CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);
	CHECK_RES_TYPE(srcmdp->args[1], ARG_STR);

	mntget(ratp[0].u.mnt);
	ratp[RAT_ROOT_MNT].count++;
	//Always check for NULL ptrs before touching them
	if(DENTRY_ARG(0) == NULL || IS_ERR(DENTRY_ARG(0)))
	{
		printk("\nreplay_lookup: Invalid dentry");
		err = -EINVAL;
        goto out;
	}
	if(ratp[srcmdp->args[1]].u.cpval == NULL || IS_ERR(ratp[srcmdp->args[1]].u.cpval))
    {
        printk("\n# replay_lookup: Invalid name");
        err = -EINVAL;
        goto out;
	}
	//printk("char ptr: %p: %s dentry ptr: %p, strlen: %d",ratp[srcmdp->args[1]].u.cpval,ratp[srcmdp->args[1]].u.cpval,DENTRY_ARG(0),strlen(ratp[srcmdp->args[1]].u.cpval));
	new_dentry = lookup_one_len(ratp[srcmdp->args[1]].u.cpval, DENTRY_ARG(0), strlen(ratp[srcmdp->args[1]].u.cpval));
	fist_print_dentry("lookup: ", new_dentry);

	CHECK_RETURN_PTR(new_dentry, srcmdp->expected_retval);

	ratp[srcmdp->ret].type	 = ARG_DENTRY;
	//ratp[srcmdp->ret].u.dentry = dget(new_dentry);

	ASSERT(atomic_read(&new_dentry->d_count) > 0);

	ratp[srcmdp->ret].u.dentry = new_dentry;
	fist_print_dentry("After adding to our cache: ", new_dentry);
	// XXX: ratp[srcmdp->ret].dentry = new_dentry;
	ratp[srcmdp->ret].count++;

out:
	print_exit_status(err);
	return err;
}

// replay unlink command
int replay_unlink(struct replay_command* srcmdp, struct replay* r)
{
	struct dentry *dentry;
	struct dentry *dir_dentry = NULL;
	struct inode *inode;
	int err = 0;
	struct res_alloc_table* ratp = r->ratp;

	print_entry_location();

	CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);

	dentry = DENTRY_ARG(0);

	dir_dentry = lock_parent(dentry);

	inode = dir_dentry->d_inode;
	fist_print_dentry("before unlink: ", dentry);
	dget(dentry);
	err = vfs_unlink(inode, dentry);
	dput(dentry);

	/*
	 * XXX: tim: below is the original implementation, unlink should not
	 * delete the dentry.  The dentry should be deleted by dput() in
	 * replay_d_release()
	*/
	/*
	if (!err) {
		fist_print_dentry("before d_delete in unlink: ", dentry);
		d_delete(dentry);
		dput(dentry);
		fist_print_dentry("after d_delete in unlink: ", dentry);
		DENTRY_ARG(0) = NULL;
		ratp[srcmdp->args[0]].type = ARG_INVAL;
	}
	*/
	unlock_dir(dir_dentry);
	CHECK_RETURN_INT(err, srcmdp->expected_retval);

out:
	print_exit_status(err);
	return err;
}

// replay open command
int replay_open(struct replay_command* srcmdp, struct replay* r)
{
	struct file *f;
	struct dentry *d;
	int err = 0;
	struct res_alloc_table* ratp = r->ratp;

	print_entry_location();

	CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);
	CHECK_RES_TYPE(srcmdp->args[1], ARG_VFSMNT);

	d = DENTRY_ARG(0);

	// TODO: see how to handle negative dentries right
	if (!d || !d->d_inode) {
		printk(KERN_WARNING"NULL dentry or d_inode in struct file\n");
		if (!d)
			printk("dentry is null\n");
		//fist_print_dentry("open:", d);
		err = -EINVAL;
		goto out;
	}

	// The only other way is to copy and paste this function :-)
	fist_print_dentry("before dentry_open", d);
	mntget(MNT_ARG(1));
	ratp[RAT_ROOT_MNT].count++;
	dget(d);
	f = dentry_open(d, MNT_ARG(1), INT_ARG(2));
	if (IS_ERR(f)) {
		dput(d);
		mntput(MNT_ARG(1));
	}
	fist_print_dentry("after dentry_open", d);

	// EAGAIN is not handled by now
	CHECK_RETURN_PTR(f, srcmdp->expected_retval);

	ratp[srcmdp->ret].type = ARG_FILE;
	ratp[srcmdp->ret].u.file = f;
	ratp[srcmdp->ret].count++;

out:
	print_exit_status(err);
	return err;
}

// replay create command
int replay_create(struct replay_command* srcmdp, struct replay* r)
{
	int err = 0;
	struct dentry *dir_dentry, *new_dentry;
	struct res_alloc_table* ratp;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	struct nameidata nd;
	char* path;
#endif
	print_entry_location();

	ratp = r->ratp;
	path = NULL;

	CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY); // I have to be created
	CHECK_RES_TYPE(srcmdp->args[1], ARG_VFSMNT);

	new_dentry = DENTRY_ARG(0);

	dir_dentry = lock_parent(new_dentry);
	err = PTR_ERR(dir_dentry);

	if (IS_ERR(dir_dentry))
		goto out_check_result;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	err = vfs_create(dir_dentry->d_inode, new_dentry, INT_ARG(2));
#else
	mntget(MNT_ARG(1));
	nd.mnt = MNT_ARG(1);
	nd.last = dir_dentry->d_name;
	nd.dentry = new_dentry;
	err = vfs_create(dir_dentry->d_inode, new_dentry, INT_ARG(2),&nd);
	mntput(MNT_ARG(1));
#endif
	unlock_dir(dir_dentry);

out_check_result:
	CHECK_RETURN_INT(err, srcmdp->expected_retval);
out:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	if(path)
	  KFREE(path);
#endif
	print_exit_status(err);
	return err;
}

/*
 * RAT[0] = OLD DENTRY
 * RAT[1] = NEW DENTRY
 */
int replay_link(struct replay_command* srcmdp, struct replay* r)
{
  int err = 0;
  struct dentry* old_dentry;
  struct dentry* new_dentry;
  struct dentry* dir_dentry;
  struct res_alloc_table* ratp;
  print_entry_location();

  ratp = r->ratp;

  CHECK_RES_TYPE(srcmdp->args[0],ARG_DENTRY);
  CHECK_RES_TYPE(srcmdp->args[1],ARG_DENTRY);

  old_dentry = DENTRY_ARG(0);
  new_dentry = DENTRY_ARG(1);
  dir_dentry = lock_parent(new_dentry);
  err = PTR_ERR(dir_dentry);

  if(IS_ERR(dir_dentry)) {
    CHECK_RETURN_INT(err,srcmdp->expected_retval);
    goto out;
  }

  err = vfs_link(old_dentry,dir_dentry->d_inode,new_dentry);

  unlock_dir(dir_dentry);
 out:

  print_exit_status(err);
  return err;
}

/*
 * RAT[0] - DENTRY
 * RAT[1] - PATHNAME
 */
int replay_symlink(struct replay_command* srcmdp,struct replay* r)
{
  int err = 0;
  struct dentry* dentry;
  struct dentry* dir_dentry;
  struct res_alloc_table* ratp;
  print_entry_location();

  ratp = r->ratp;

  CHECK_RES_TYPE(srcmdp->args[0],ARG_DENTRY);
  CHECK_RES_TYPE(srcmdp->args[1],ARG_STR);

  dentry = DENTRY_ARG(0);
  dir_dentry = lock_parent(dentry);
  err = PTR_ERR(dir_dentry);

  if(IS_ERR(dir_dentry)) {
    CHECK_RETURN_INT(err,srcmdp->expected_retval);
    goto out;
  }

  err = vfs_symlink(dir_dentry->d_inode,dentry,STR_ARG(1),0);

  unlock_dir(dir_dentry);
 out:

  print_exit_status(err);
  return err;
}

/*
 * RAT[0] - OLD DENTRY
 * RAT[1] - NEW DENTRY
 */
int replay_rename(struct replay_command* srcmdp,struct replay* r)
{
  int err = 0;
  struct dentry* old_dentry;
  struct dentry* old_dir_dentry;
  struct dentry* new_dentry;
  struct dentry* new_dir_dentry;
  struct res_alloc_table* ratp;
  print_entry_location();

  ratp = r->ratp;

  CHECK_RES_TYPE(srcmdp->args[0],ARG_DENTRY);
  CHECK_RES_TYPE(srcmdp->args[1],ARG_DENTRY);

  old_dentry = DENTRY_ARG(0);
  new_dentry = DENTRY_ARG(1);

  old_dir_dentry = lock_parent(old_dentry);
  if(IS_ERR(old_dir_dentry)) {
    CHECK_RETURN_INT(err,srcmdp->expected_retval);
    goto out;
  }

  if(new_dentry->d_parent != old_dir_dentry) {
    new_dir_dentry = lock_parent(new_dentry);
    if(IS_ERR(new_dir_dentry)) {
      unlock_dir(old_dir_dentry);
      CHECK_RETURN_INT(err,srcmdp->expected_retval);
      goto out;
    }
  }
  else
    new_dir_dentry = old_dir_dentry; //Same parent

  err = vfs_rename(old_dir_dentry->d_inode,old_dentry,
		   new_dir_dentry->d_inode,new_dentry);

  unlock_dir(old_dir_dentry);
  if(new_dentry->d_parent != old_dentry->d_parent)
    unlock_dir(new_dir_dentry);
 out:
  print_exit_status(err);
  return err;
}

int replay_load(struct replay_command* srcmdp, struct replay* r)
{
	int err = 0;
	mm_segment_t old_fs;
	int length;
	loff_t offset;
	struct res_alloc_table* ratp = r->ratp;
	struct file* data = r->data;

	print_entry_location();

	if(!data || !data->f_op || !data->f_op->read) {
		err = -EINVAL;
		goto out;
	}

	offset = srcmdp->args[0];
	length = srcmdp->args[1];

	ratp[srcmdp->ret].u.cpval = VMALLOC(length);
	if (!ratp[srcmdp->ret].u.cpval) {
		err = -ENOMEM;
		goto out;
	}

	old_fs = get_fs();
	set_fs(KERNEL_DS);
	err = data->f_op->read(data, ratp[srcmdp->ret].u.cpval, length, &offset);
	set_fs(old_fs);

	if (err < length) {
		if (err >= 0)
			err = -EIO;
		goto out;
	}
	err = 0;

	ratp[srcmdp->ret].type = ARG_STR;

out:
	print_exit_status(err);
	if (err < 0 && ratp[srcmdp->ret].u.cpval) {
	  VFREE(ratp[srcmdp->ret].u.cpval);
	  ratp[srcmdp->ret].u.cpval = NULL;
	  ratp[srcmdp->ret].type = ARG_INVAL;
	}
	return err;
}

int replay_unload(struct replay_command* srcmdp, struct replay* r)
{
	int err = 0;
	struct res_alloc_table* ratp = r->ratp;

	print_entry_location();

	CHECK_RES_TYPE(srcmdp->args[0], ARG_STR);
	VFREE(STR_ARG(0));

	STR_ARG(0) = NULL;
	ratp[srcmdp->args[0]].type = ARG_INVAL;
out:
	print_exit_status(err);
	return err;
}

int replay_prefetch_load(struct replay_command* srcmdp, struct replay* r)
{
	int err = 0;
	mm_segment_t old_fs;
	int length;
	loff_t offset;
	struct res_alloc_table* ratp = r->ratp;
	struct file* data = r->data;

	print_entry_location();

	if(!data || !data->f_op || !data->f_op->read) {
		err = -EINVAL;
		goto out;
	}

	offset = srcmdp->args[0];
	length = srcmdp->args[1];

	ratp[srcmdp->ret].u.cpval = prefetch_get_buffer(data,offset,length);

	if(!ratp[srcmdp->ret].u.cpval) {
	  err = -ENOMEM;
	  goto out;
	}

	ratp[srcmdp->ret].type = ARG_STR;

out:
	print_exit_status(err);
	return err;
}

int replay_prefetch_unload(struct replay_command* srcmdp, struct replay* r)
{
  int err = 0;
  long offset;
  struct res_alloc_table* ratp = r->ratp;

  print_entry_location();

  offset = INT_ARG(1);
  prefetch_release_buffer(r->data,STR_ARG(0),offset);
  STR_ARG(0) = NULL;
  ratp[srcmdp->args[0]].type = ARG_INVAL;

  print_exit_location();
  return err;
}

#if MULTITHREADING
int replay_create_thread(struct replay_command* srcmdp,struct replay *r)
{
  int err = 0;
  struct semaphore* sem;
  struct res_alloc_table* ratp = r->ratp;
  kthread_data_t *kthread_data;
  print_entry_location();

  //Create the kthread_data object
  kthread_data = KMALLOC(sizeof(kthread_data_t),GFP_KERNEL);
  if(!kthread_data) {
    print_error_nomem();
    err = -ENOMEM;
    goto out;
  }
  memset(kthread_data,0,sizeof(kthread_data_t));

  //Create the semaphore
  sem = KMALLOC(sizeof(struct semaphore),GFP_KERNEL);
  if(!sem) {
    print_error_nomem();
    err = -ENOMEM;
    goto out;
  }
  kthread_data->suspend = KMALLOC(sizeof(atomic_t),GFP_KERNEL);
  if(!kthread_data->suspend) {
    print_error_nomem();
    err = -ENOMEM;
    goto out;
  }
  atomic_set(kthread_data->suspend,0);
  sema_init(sem,1);
  kthread_data->sem = sem;
  kthread_data->srcmdp = srcmdp;
  kthread_data->r = r;
  kthread_data->thread_loc = srcmdp->ret;
  kthread_data->terminate = 0;
  kthread_data->initialized = 0;

  //Start the kthread
  if(init_thread_loc)
    kthread_data->task = kthread_run(thread_execute,kthread_data,"Replayfs-Helper");
  else {
    kthread_data->task = current;
    init_thread_loc = srcmdp->ret;
  }

  ratp[srcmdp->ret].u.kthread_data = kthread_data;
  ratp[srcmdp->ret].type = ARG_KTHREAD_DATA;

 out:
  print_exit_location();
  return err;
}
#endif

int replay_d_release(struct replay_command* srcmdp, struct replay *r)
{
	int err = 0;
	struct res_alloc_table* ratp = r->ratp;

	print_entry_location();

	CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);

	//The ref. count must be 1, otherwise there MAY be a bug somewhere
	if(atomic_read(&DENTRY_ARG(0)->d_count) != 1) {
	  printk("WARNING: D_RELEASE when refcount is greater than one - d_count:%d\n",atomic_read(&DENTRY_ARG(0)->d_count));
	}
	// Let go of this dentry, the cache wants to get cleaned!
	dput(DENTRY_ARG(0));

	ratp[srcmdp->args[0]].type = ARG_INVAL;
	DENTRY_ARG(0) = NULL;
out:
	print_exit_status(err);
	return err;
}

extern page_t* replay_get1page(file_t* filp, int index);

int replay_readpage(struct replay_command* srcmdp, struct replay* r)
{
	int err;
	struct file *f;
	page_t* page;
	struct res_alloc_table* ratp;

	print_entry_location();

	err = 0;
	ratp = r->ratp;
	page = NULL;

	CHECK_RES_TYPE(srcmdp->args[0], ARG_FILE);
	//This can be 0 if read capturing was disabled
	if(srcmdp->args[1])
	  CHECK_RES_TYPE(srcmdp->args[1], ARG_STR);

	f = FILE_ARG(0);

	/*buffer = KMALLOC(PAGE_SIZE, GFP_KERNEL);
	if (!buffer) {
		err = -ENOMEM;
		goto out;
	}*/

	/* TODO: force read even if page is in the cache */
	page = replay_get1page(f, INT_ARG(2));
	if (IS_ERR(page)) {
		err = PTR_ERR(page);
		goto out;
	}

	CHECK_RETURN_INT(err, srcmdp->expected_retval);

out:
	if (page)
		page_cache_release(page);
	print_exit_status(err);
	return err;
}

//XXX: untested
int replay_readpages(struct replay_command* srcmdp, struct replay* r)
{
	int err;
	struct file *f;
	struct res_alloc_table* ratp;
	struct address_space* mapping;
	struct list_head list;
	struct list_head* lptr;
	int pos;
	char* buffer;
	long page_index;
	long page_buffer_rat_loc;
	page_t* page;
	page_t* first_page;
	int i;

	print_entry_location();

	err = 0;
	ratp = r->ratp;
	first_page = NULL;

	CHECK_RES_TYPE(srcmdp->args[0], ARG_FILE);
	CHECK_RES_TYPE(srcmdp->args[2], ARG_STR);

	f = FILE_ARG(0);
	mapping = f->f_dentry->d_inode->i_mapping;

	//If lower file system does not support readpages() then fail!
	if(mapping->a_ops->readpages == NULL) {
	  err = -ENOSYS;
	  goto out;
	}

	buffer = STR_ARG(2);

	//Setup the list page
	INIT_LIST_HEAD(&list);
	for(i = 0,pos = 0;i < INT_ARG(1);i++) {
	  memcpy(&page_index,buffer+pos,sizeof(long));
	  pos += sizeof(long);
	  memcpy(&page_buffer_rat_loc,buffer+pos,sizeof(long));
	  pos += sizeof(long);

	  page = find_or_create_page(mapping,page_index,GFP_KERNEL);
	  unlock_page(page);

	  list_add(&page->lru,&list);
	}

	err = mapping->a_ops->readpages(f,mapping,&list,INT_ARG(1));

	CHECK_RETURN_INT(err, srcmdp->expected_retval);

	//Release pages, assume circular list
	//XXX: tim: might not be necessary
	list_for_each(lptr,&list) {
	  page = list_entry(lptr,page_t,lru);
	  if(page == first_page)
	    break;
	  else if(first_page == NULL)
	    first_page = page;
	  page_cache_release(page);
	}

	// TODO: compare contents of the corresponding buffers
 out:
	print_exit_status(err);
	return err;
}

/*
 * ARG[0] = ARG_DENTRY
 * ARG[1] = ARG_STR
 * ARG[2] = INT
 */
int replay_writepage(struct replay_command* srcmdp, struct replay* r)
{
	int err = 0;
	loff_t offset;
	struct res_alloc_table* ratp=NULL;
	struct inode* inode=NULL;
	struct page* page=NULL;
	struct pagevec* pvec;

	struct writeback_control wbc = {
	  .bdi = NULL,
	  .sync_mode = WB_SYNC_ALL,
	};
	//Always check for NULL before touching them
	if(r == NULL){
	    err = -1;
	    printk("\nError: replay NULL");
	    goto out;
	}
	if(srcmdp == NULL){
	  err = -1;
	  printk("\nError: replaycommand NULL");
	  goto out;
	}

	print_entry_location();

	err = 0;
	ratp = r->ratp;

	CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);
	//CHECK_RES_TYPE(srcmdp->args[1], ARG_STR);

	inode = DENTRY_ARG(0)->d_inode;
	if(!inode || !(inode->i_mapping)){
	   printk("Error: NULL inode read from DENTRY_ARG(0)\n");
	   err = -1;
	   goto out;
	}
	//This print will be retained for some more time until writepage gets fully mature
 	printk("replay_writepage inum %ld",inode->i_ino);
	page = find_or_create_page(inode->i_mapping,INT_ARG(2),GFP_KERNEL);

	if(!page || IS_ERR(page)){
	    printk("Error: inum: %ld NULL page read from find_or_createpage\n",inode->i_ino);
	    err = -1;
	    goto out;
	}

	wait_on_page_writeback(page);

	if(!(page->mapping->a_ops->writepage) || IS_ERR(page->mapping->a_ops->writepage)){
	  printk("Error: No aop->writepage for inum %ld\n",inode->i_ino);
          err = -1;
          goto out;
	}
	page->mapping->a_ops->writepage(page,&wbc);
	//Writepage unlocks the page for us

	page_cache_release(page);
	CHECK_RETURN_INT(err, srcmdp->expected_retval);
out:
	print_exit_status(err);
	return err;
}

int replay_dget(struct replay_command* srcmdp, struct replay* r)
{
  int err = 0;
  struct res_alloc_table* ratp;
  struct dentry* dentry;

  print_entry_location();

  ratp = r->ratp;
  CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);
  dentry = DENTRY_ARG(0);
  ASSERT(atomic_read(&dentry->d_count) > 0);
  dget(dentry);

 out:
  print_exit_status(err);
  return err;
}

int replay_dput(struct replay_command* srcmdp, struct replay* r)
{
  int err = 0;
  struct res_alloc_table* ratp;
  struct dentry* dentry;

  print_entry_location();

  ratp = r->ratp;
  CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);
  dentry = DENTRY_ARG(0);

  ASSERT(atomic_read(&dentry->d_count) > 0);
  dput(dentry);

 out:
  print_exit_status(err);
  return err;
}

/*
 * ARG[0] = ARG_DENTRY
 * ARG[1] = INT
 * ARG[2] = ULONG
 */
int replay_mknod(struct replay_command *srcmdp, struct replay *r)
{
  int err = 0;
  struct res_alloc_table* ratp;
  struct dentry* dentry;
  struct dentry* dir_dentry;
  print_entry_location();

  ratp = r->ratp;
  CHECK_RES_TYPE(srcmdp->args[0],ARG_DENTRY);
  dentry = DENTRY_ARG(0);
  ASSERT(dentry);

  dir_dentry = lock_parent(dentry);
  ASSERT(dir_dentry);

  err = vfs_mknod(dir_dentry->d_inode,
		  dentry,
		  INT_ARG(3),
		  INT_ARG(2)); /* XXX: tim: assuming 32-bit system - incorrect */

  if(dir_dentry)
    unlock_dir(dir_dentry);

  CHECK_RETURN_INT(err,srcmdp->expected_retval);

 out:
  print_exit_status(err);
  return err;
}

/*
 * ARG[0] = ARG_DENTRY
 * ARG[1] = ARG_STR (struct iattr)
 */
int replay_setattr(struct replay_command *srcmdp, struct replay *r)
{
  int err = 0;
  struct res_alloc_table* ratp;
  struct dentry* dentry;
  struct iattr* ia;
  print_entry_location();

  ratp = r->ratp;
  CHECK_RES_TYPE(srcmdp->args[0],ARG_DENTRY);
  CHECK_RES_TYPE(srcmdp->args[1],ARG_STR);
  dentry = DENTRY_ARG(0);
  ia = (struct iattr*)STR_ARG(1);

  /*
   * XXX: tim: not going to truncate, assume that another replay op will take
   * care of it
   */

  err = notify_change(dentry,ia);

  CHECK_RETURN_INT(err,srcmdp->expected_retval);
 out:
  print_exit_status(err);
  return err;
}

/*
 * ARG[0] = ARG_DENTRY
 * ARG[1] = ARG_DENTRY
 * ARG[2] = ARG_INT_FLAGS
 */
int replay_fsync(struct replay_command *srcmdp, struct replay *r)
{
  int err = 0;
  struct res_alloc_table* ratp;
  struct dentry *dentry;
  struct file *file;
  int datasync;
  print_entry_location();

  ratp = r->ratp;
  CHECK_RES_TYPE(srcmdp->args[0],ARG_DENTRY);
  CHECK_RES_TYPE(srcmdp->args[1],ARG_FILE);

  dentry = DENTRY_ARG(0);
  file = FILE_ARG(1);
  datasync = INT_ARG(2);

  if(dentry->d_inode->i_fop && dentry->d_inode->i_fop->fsync)
    err = dentry->d_inode->i_fop->fsync(file,dentry,datasync);
  else if(file->f_op && file->f_op->fsync)
    err = file->f_op->fsync(file,dentry,datasync);
  else
    err = -ENOSYS;

  CHECK_RETURN_INT(err,srcmdp->expected_retval);
 out:
  print_exit_status(err);
  return err;
}

int replay_llseek(struct replay_command *srcmdp,struct replay *r)
{
  int err = 0;
  struct res_alloc_table *ratp;
  struct file *file;
  loff_t *offset;
  int origin;
  print_entry_location();

  ratp = r->ratp;
  CHECK_RES_TYPE(srcmdp->args[0],ARG_FILE);
  CHECK_RES_TYPE(srcmdp->args[1],ARG_STR);

  file = FILE_ARG(0);
  offset = (loff_t*)STR_ARG(1);
  origin = INT_ARG(2);

  ASSERT(file);
  if(file->f_op && file->f_op->llseek)
    file->f_op->llseek(file,*offset,origin);
  else
    generic_file_llseek(file,*offset,origin);

 out:
  print_exit_status(err);
  return err;
}

int replay_permission(struct replay_command *srcmdp,struct replay *r)
{
  int err = 0;
  struct res_alloc_table *ratp;
  struct nameidata nd;
  struct dentry *dentry;
  int mask;
  print_entry_location();

  ratp = r->ratp;
  CHECK_RES_TYPE(srcmdp->args[0],ARG_DENTRY);
  CHECK_RES_TYPE(srcmdp->args[1],ARG_VFSMNT);

  dentry = DENTRY_ARG(0);
  nd.mnt = MNT_ARG(1);
  mask = INT_ARG(2);

  nd.last = dentry->d_parent->d_name;
  nd.dentry = dentry;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
  vfs_permission(dentry->d_inode,mask);
#else
  permission(dentry->d_inode,mask,&nd);
#endif

 out:
  print_exit_status(err);
  return err;
}

int replay_commit_write(struct replay_command *srcmdp, struct replay *r)
{
	unsigned long index;
	struct inode *inode;
	struct res_alloc_table* ratp = r->ratp;
	struct file *file;
	//char *buffer = NULL;
	page_t *page=NULL;
	int err = 0;

	print_entry_location();
	if(!srcmdp ){
        err = -EINVAL;
        goto out;
    }

	CHECK_RES_TYPE(srcmdp->args[0], ARG_FILE);
	CHECK_RES_TYPE(srcmdp->args[1], ARG_STR);
	file = FILE_ARG(0);
	if(!file || IS_ERR(file)){
        err = -EINVAL;
        goto out;
	}
	inode = file->f_dentry->d_inode;
    if(!inode || IS_ERR(inode)){
        err = -EINVAL;
        goto out;
    }

	index = INT_ARG(2);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	down(&inode->i_sem);
#else
	mutex_lock(&inode->i_mutex);
#endif
	if(!inode->i_mapping || IS_ERR(inode->i_mapping)){
        err = -EINVAL;
        goto out_unlock;
    }

	page = grab_cache_page(inode->i_mapping, index);

	if (NULL == page || IS_ERR(page)){
		err = -EINVAL;
		goto out_check_return;
	}

	err = -EINVAL;
	if (inode->i_mapping &&
	    inode->i_mapping->a_ops &&
	    inode->i_mapping->a_ops->prepare_write){
		err = inode->i_mapping->a_ops->prepare_write(file, page, INT_ARG(3), INT_ARG(4));
	}

	if (err < 0)
	  goto out_unlock;
	ASSERT(R_INT_ARG(1) != 0);

	if(NULL == page_address(page)){
	/*	err = -EINVAL;
		goto out_unlock;*/
	}
	else
		memcpy((char *)page_address(page) + INT_ARG(3), (char *)R_INT_ARG(1) + INT_ARG(3), INT_ARG(4) - INT_ARG(3));

	wait_on_page_writeback(page);
	err = inode->i_mapping->a_ops->commit_write(file, page, INT_ARG(3), INT_ARG(4));

	if(err) //Failed to commit_write()
	  goto out_unlock;

	/*
	 * Contrived error condition.  tracefs_commit_write() was also  modified
	 * to return what a_ops->commit_write() returns rather than the number
	 * of bytes VFS was SUPPOSED to write.
	 */
	//err = INT_ARG(4) - INT_ARG(3);

out_unlock:
	unlock_page(page);
	page_cache_release(page);

out_check_return:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	up(&inode->i_sem);
#else
	mutex_unlock(&inode->i_mutex);
#endif
	CHECK_RETURN_INT(err, srcmdp->expected_retval);

out:

	print_exit_status(err);
	return err;
}

int replay_sync_page(struct replay_command* srcmdp,struct replay* r)
{
  int err;
  struct res_alloc_table* ratp;
  struct inode *inode;
  page_t* page;

  print_entry_location();

  err = 0;
  ratp = r->ratp;

  CHECK_RES_TYPE(srcmdp->args[0],ARG_DENTRY);

  inode = DENTRY_ARG(0)->d_inode;
  page = find_get_page(inode->i_mapping,INT_ARG(2));
  //Check for NULL pointers before touching them
  if(!page || IS_ERR(page)) {
  //Do not return any error here. This is not an error condition. It only
  //means that there was no page found for this inode as it was already flushed
  //by the pdflush daemon.
  //  err = -ENOENT;
    goto out;
  }

  wait_on_page_writeback(page);
  if(page->mapping && page->mapping->a_ops && page->mapping->a_ops->sync_page)
    page->mapping->a_ops->sync_page(page);

 out:
  print_exit_location();
  return err;
}

int replay_mkdir(struct replay_command *srcmdp, struct replay *r)
{
	int err;
	struct res_alloc_table *ratp;
	struct dentry *dir, *dentry;

	print_entry_location();

	err = 0;
	ratp = r->ratp;

	CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);

	dentry = DENTRY_ARG(0);
	dir = lock_parent(dentry);

	if (IS_ERR(dir)) {
		err = PTR_ERR(dir);
		goto  out_check_result;
	}
	err = vfs_mkdir(dir->d_inode, dentry, INT_ARG(2));

	unlock_dir(dir);
	if (!dentry->d_inode)
		d_drop(dentry);

out_check_result:
	CHECK_RETURN_INT(err, srcmdp->expected_retval);

out:
	print_exit_status(err);

	return err;
}

int replay_rmdir(struct replay_command *srcmdp, struct replay *r)
{
	int err;
	struct res_alloc_table *ratp;
	struct dentry *dir, *dentry;

	print_entry_location();

	err = 0;
	ratp = r->ratp;

	CHECK_RES_TYPE(srcmdp->args[0], ARG_DENTRY);

	dentry = DENTRY_ARG(0);
	dir = lock_parent(dentry);

	if (IS_ERR(dir)) {
		err = PTR_ERR(dir);
		goto  out_check_result;
	}
	dget(dentry);
	err = vfs_rmdir(dir->d_inode, dentry);
	dput(dentry);

	if (!err)
		d_delete(dentry);
	unlock_dir(dir);

out_check_result:

	CHECK_RETURN_INT(err, srcmdp->expected_retval);

out:
	print_exit_status(err);

	return err;
}


struct replay_readdir_callback {
	int count;
};

/*
 * copied from fs/readdir.c
 */
struct linux_dirent64 {
	u64             d_ino;
	s64             d_off;
	unsigned short  d_reclen;
	unsigned char   d_type;
	char            d_name[0];
};
#define NAME_OFFSET(de) ((int) ((de)->d_name - (char *) (de)))
#define ROUND_UP64(x) (((x)+sizeof(u64)-1) & ~(sizeof(u64)-1))

static int replay_filldir(void *buf, const char *name, int namlen, loff_t offset, ino_t ino, unsigned d_type)
{
	struct linux_dirent64 *dirent;
	int reclen = ROUND_UP64(NAME_OFFSET(dirent) + namlen + 1);
	struct replay_readdir_callback *cb = (struct replay_readdir_callback *)buf;

	if (reclen > cb->count)
		return -EINVAL;

	cb->count -= reclen;
	return 0;
}

int replay_readdir(struct replay_command *srcmdp, struct replay *r)
{
	int err = 0;
	struct res_alloc_table *ratp = r->ratp;
	struct replay_readdir_callback cb;


	print_entry_location();

	CHECK_RES_TYPE(srcmdp->args[0], ARG_FILE);

	cb.count = INT_ARG(2) - INT_ARG(3);

	err = vfs_readdir(FILE_ARG(0), replay_filldir, (void *)&cb);

	CHECK_RETURN_INT(err, srcmdp->expected_retval);
out:

	print_exit_status(err);

	return err;

}
