/*
 * blowfish cipher
 */
/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 */

//#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <asm/byteorder.h>
#include "crypto.h"
#include "blowfish.h"


inline void
_blowfish_encrypt_block (const u32 in[2], u32 out[2], const blowfish_key_t *key)
{
	const u32 *P = key->P;
	const u32 *S = key->S;

	register u32 yl = in[0];
	register u32 yr = in[1];

	ROUND (yr, yl, 0);
	ROUND (yl, yr, 1);
	ROUND (yr, yl, 2);
	ROUND (yl, yr, 3);
	ROUND (yr, yl, 4);
	ROUND (yl, yr, 5);
	ROUND (yr, yl, 6);
	ROUND (yl, yr, 7);
	ROUND (yr, yl, 8);
	ROUND (yl, yr, 9);
	ROUND (yr, yl, 10);
	ROUND (yl, yr, 11);
	ROUND (yr, yl, 12);
	ROUND (yl, yr, 13);
	ROUND (yr, yl, 14);
	ROUND (yl, yr, 15);

	/* yl and yr are switched */
	yl ^= P[16];
	yr ^= P[17];

	out[0] = yr;
	out[1] = yl;
}

int
blowfish_encrypt (struct tf_cipher_context *cx,
		  const u8 *in8, u8 *out8, int size, int atomic)
{
	const blowfish_key_t *key = (blowfish_key_t *) cx->keyinfo;
	const u32 *in_blk = (const u32 *) in8;
	u32 * const out_blk = (u32 *) out8;
	u32 _in32[2], _out32[2];

	if (size != 8)
		return 1;

	_in32[0] = be32_to_cpu (in_blk[0]);
	_in32[1] = be32_to_cpu (in_blk[1]);

	_blowfish_encrypt_block (_in32, _out32, key);

	out_blk[0] = cpu_to_be32 (_out32[0]);
	out_blk[1] = cpu_to_be32 (_out32[1]);

	return 0;
}

int
blowfish_decrypt (struct tf_cipher_context *cx,
		  const u8 *in8, u8 *out8, int size, int atomic)
{
	const blowfish_key_t *key = (blowfish_key_t *) cx->keyinfo;
	const u32 *in_blk = (const u32 *) in8;
	u32 * const out_blk = (u32 *) out8;

	const u32 *P = key->P;
	const u32 *S = key->S;

	if (size != 8)
		return 1;

	{
		register u32 yl = be32_to_cpu (in_blk[0]);
		register u32 yr = be32_to_cpu (in_blk[1]);

		ROUND (yr, yl, 17);
		ROUND (yl, yr, 16);
		ROUND (yr, yl, 15);
		ROUND (yl, yr, 14);
		ROUND (yr, yl, 13);
		ROUND (yl, yr, 12);
		ROUND (yr, yl, 11);
		ROUND (yl, yr, 10);
		ROUND (yr, yl, 9);
		ROUND (yl, yr, 8);
		ROUND (yr, yl, 7);
		ROUND (yl, yr, 6);
		ROUND (yr, yl, 5);
		ROUND (yl, yr, 4);
		ROUND (yr, yl, 3);
		ROUND (yl, yr, 2);

		/* yl and yr are switched */
		yl ^= P[1];
		yr ^= P[0];

		out_blk[0] = cpu_to_be32 (yr);
		out_blk[1] = cpu_to_be32 (yl);
	}
	return 0;
}

/* Sets the blowfish S and P boxes for encryption and decryption. */

int
blowfish_set_key (struct tf_cipher_context *cx,
		  const u8 *key, int key_len, int atomic)
{
	blowfish_key_t *key2 = (blowfish_key_t *) cx->keyinfo;
	short i;
	short j;
	short count;
	u32 data[2];
	u32 temp;
	u32 *P = key2->P;
	u32 *S = key2->S;

	/* Check key length. */

	if (key_len != 8 && key_len != 16 && key_len != 20
	    && key_len != 24 && key_len != 32)
		return -EINVAL;	/* unsupported key length */

	cx->key_length = key_len;


	/* Copy the initialization s-boxes */

	for (i = 0, count = 0; i < 256; i++)
		for (j = 0; j < 4; j++, count++)
			S[count] = bf_sbox[count];

	/* Set the p-boxes */

	for (i = 0; i < 16 + 2; i++)
		P[i] = bf_pbox[i];

	/* Actual subkey generation */

	for (j = 0, i = 0; i < 16 + 2; i++) {
		temp = (((u32) key[j] << 24) |
			((u32) key[(j + 1) % key_len] << 16) |
			((u32) key[(j + 2) % key_len] << 8) |
			((u32) key[(j + 3) % key_len]));

		P[i] = P[i] ^ temp;
		j = (j + 4) % key_len;
	}

	data[0] = 0x00000000;
	data[1] = 0x00000000;

	for (i = 0; i < 16 + 2; i += 2) {
		_blowfish_encrypt_block (data, data, key2);

		P[i] = data[0];
		P[i + 1] = data[1];
	}

	for (i = 0; i < 4; i++) {
		for (j = 0, count = i * 256; j < 256; j += 2, count += 2) {
			_blowfish_encrypt_block (data, data, key2);

			S[count] = data[0];
			S[count + 1] = data[1];
		}
	}
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_NO_SYMBOLS;
#endif

/* eof */
