/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: mmap.c,v 1.48 2007/11/06 23:38:30 gopalas Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_tracefs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "tracefs.h"
#include "tracer.h"

STATIC int
tracefs_prepare_write(file_t *file, page_t *page, unsigned from, unsigned to);
STATIC int
tracefs_commit_write(file_t *file, page_t *page, unsigned from, unsigned to);

/*
 *  * Return byte-offset into filesystem object for page.
 *   */
static inline loff_t tracefs_page_offset(struct page *page)
{
	return ((loff_t)page->index) << PAGE_CACHE_SHIFT;
}

/*
 * get one page from cache or lower f/s, return error otherwise.
 * returns unlocked, up-to-date page (if ok), with increased refcnt.
 */
page_t *
tracefs_get1page(file_t *file, int index)
{
    page_t *page;
    dentry_t *dentry;
    inode_t *inode;
    struct address_space *mapping;
    int err;

    print_entry_location();

    dentry = file->f_dentry; /* CPW: Moved below print_entry_location */
    inode = dentry->d_inode;
    mapping = inode->i_mapping;

    fist_dprint(8, "%s: read page index %d pid %d\n", __FUNCTION__, index, current->pid);
    if (index < 0) {
	printk("%s BUG: index=%d\n", __FUNCTION__, index);
	page = ERR_PTR(-EIO);
	goto out;
    }
    page = read_cache_page(mapping,
			   index,
			   (filler_t *) mapping->a_ops->readpage,
			   (void *) file);
    if (IS_ERR(page))
	goto out;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    wait_on_page(page);
    if (!Page_Uptodate(page)) {
#else
    wait_on_page_locked(page);
    if (!PageUptodate(page)) {
#endif
	lock_page(page);
	err = mapping->a_ops->readpage(file, page);
	if (err) {
	    page = ERR_PTR(err);
	    goto out;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	wait_on_page(page);
	if (!Page_Uptodate(page)) {
#else
	wait_on_page_locked(page);
	if (!PageUptodate(page)) {
#endif
	    page = ERR_PTR(-EIO);
	    goto out;
	}
    }

 out:
    print_exit_pointer(page);
    return page;
}


/*
 * get one page from cache or lower f/s, return error otherwise.
 * similar to get1page, but doesn't guarantee that it will return
 * an unlocked page.
 */
page_t *
tracefs_get1page_cached(file_t *file, int index)
{
    page_t *page;
    dentry_t *dentry;
    inode_t *inode;
    struct address_space *mapping;

    print_entry_location();

    dentry = file->f_dentry; /* CPW: Moved below print_entry_location */
    inode = dentry->d_inode;
    mapping = inode->i_mapping;

    fist_dprint(8, "%s: read page index %d pid %d\n", __FUNCTION__, index, current->pid);
    if (index < 0) {
	printk("%s BUG: index=%d\n", __FUNCTION__, index);
	page = ERR_PTR(-EIO);
	goto out;
    }
    page = read_cache_page(mapping,
			   index,
			   (filler_t *) mapping->a_ops->readpage,
			   (void *) file);
    if (IS_ERR(page))
	goto out;

 out:
    print_exit_pointer(page);
    return page;
}

/*
 * Function for handling creation of holes when lseek-ing past the
 * end of the file and then writing some data.
 */
int
tracefs_fill_zeros(file_t* file, page_t *page, unsigned from)
{
    int err = 0;
    dentry_t *dentry = file->f_dentry;
    inode_t *inode = dentry->d_inode;
    page_t *tmp_page;
    int index;

    print_entry_location();
    for (index = i_size_read(inode) >> PAGE_CACHE_SHIFT; index < page->index; index++) {
	tmp_page = tracefs_get1page(file, index);
	if (IS_ERR(tmp_page)) {
	    err = PTR_ERR(tmp_page);
	    goto out;
	}
	kmap(tmp_page);
	/*
	 * zero out rest of the contents of the page between the appropriate
	 * offsets.
	 */
	memset((char*)page_address(tmp_page) + (i_size_read(inode) & ~PAGE_CACHE_MASK), 0, PAGE_CACHE_SIZE - (i_size_read(inode) & ~PAGE_CACHE_MASK));

	if (! (err = tracefs_prepare_write(file, tmp_page, 0, PAGE_CACHE_SIZE))) {
	    wait_on_page_writeback(tmp_page);
            err = tracefs_commit_write(file, tmp_page, 0, PAGE_CACHE_SIZE);
	}

	kunmap(tmp_page);
	page_cache_release(tmp_page);
	if (err < 0)
		goto out;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        if (current->need_resched)
            schedule();
#else
	if(need_resched())
	    schedule();
#endif
    }

    /* zero out appropriate parts of last page */


    if ((from - (i_size_read(inode) & ~PAGE_CACHE_MASK)) > 0) {

	memset((char*)page_address(page) + (i_size_read(inode) & ~PAGE_CACHE_MASK), 0, from - (i_size_read(inode) & ~PAGE_CACHE_MASK));
	if (! (err = tracefs_prepare_write(file, page, 0, PAGE_CACHE_SIZE)))
            err = tracefs_commit_write(file, page, 0, PAGE_CACHE_SIZE);

	if (err < 0)
		goto out;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        if (current->need_resched)
            schedule();
#else
	if(need_resched())
	    schedule();
#endif
    }
out:
    print_exit_status(err);
    return err;
}

STATIC int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
tracefs_writepage(page_t *page)
#else
tracefs_writepage(page_t *page,struct writeback_control *wbc)
#endif
{
    int err = -EIO;
    inode_t *inode;
    inode_t *hidden_inode;
    void *context;

    print_entry_location();

    inode = page->mapping->host;
    trace_writepage_pre(stoptr(inode->i_sb), page, &context);
    hidden_inode = itohi(inode);

    page->mapping = hidden_inode->i_mapping;

    /*
     * XXX: tim: We MUST NOT call the lower-fs' writepage if the
     * writeback flag is still set on the lower-fs' page, since with
     * no double-caching there is no lower page, mpage_writepage() should
     * be handling it.
     */
    ASSERT(!PageWriteback(page));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    err = hidden_inode->i_mapping->a_ops->writepage(page);
#else
    err = hidden_inode->i_mapping->a_ops->writepage(page,wbc);
#endif
    page->mapping = inode->i_mapping;
    /*
     * update mtime and ctime of lower level file system
     * tracefs' mtime and ctime are updated by generic_file_write
     */
    hidden_inode->i_mtime = hidden_inode->i_ctime = CURRENT_TIME;

    trace_writepage_post(stoptr(inode->i_sb), page, err, &context);
    print_exit_status(err);
    return err;
}

//XXX: FIX ME!!!
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int tracefs_readpages(struct file *file, struct address_space *mapping,
			     struct list_head *pages, unsigned nr_pages)
{
    int err  = 0;
    file_t *hidden_file = NULL;
    inode_t *inode;
    inode_t *hidden_inode;
    struct address_space* hidden_mapping;
    void* context;
    struct list_head* list;
    page_t *page;

    print_entry_location();

    inode = file->f_dentry->d_inode;
    hidden_inode = itohi(inode);
    trace_readpages_pre(stoptr(inode->i_sb),file,mapping,pages,&context);
    if(ftopd(file) != NULL)
	hidden_file = ftohf(file);
    hidden_mapping = hidden_inode->i_mapping;

    //Write to the lower file system
    /*
     * XXX: tim: problem with this, this succeeds, but other operations
     * fail afterwards.  The mapping for each page cannot be flipped to
     * point to the lower file system since page->mapping = NULL for
     * all pages in the list.
     */
    err = hidden_mapping->a_ops->readpages(hidden_file,hidden_mapping,
					   pages,nr_pages);

    trace_readpages_post(stoptr(inode->i_sb),file,mapping,pages,err,&context);

    print_exit_status(err);
    return err;
}
#endif //#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)


STATIC int
tracefs_readpage(file_t *file, page_t *page)
{
	int err = -EIO;
	void *context;
	file_t *hidden_file = NULL;
	inode_t *inode;
	inode_t *hidden_inode;
	loff_t offset;
	mm_segment_t old_fs;
	char *page_data = NULL;

	print_entry_location();
	trace_readpage_pre(stoptr(file->f_dentry->d_inode->i_sb), file, page, &context);

	if (ftopd(file) != NULL)
	hidden_file = ftohf(file);

#ifdef TRACEFS_MMAP_CHANGES
	tracefs_read_lock(file->f_dentry->d_sb);
	page_data = (char *) kmap(page);
	/*
	 * Use vfs_read because some lower file systems don't have a
	 * readpage method, and some file systems (esp. distributed ones)
	 * don't like their pages to be accessed directly.  Using vfs_read
	 * may be a little slower, but a lot safer, as the VFS does a lot of
	 * the necessary magic for us.
	 */

	hidden_file->f_pos = tracefs_page_offset(page);
	offset = tracefs_page_offset(page);
	old_fs = get_fs();
	set_fs(KERNEL_DS);
	err = vfs_read(hidden_file, page_data, PAGE_CACHE_SIZE,
        	       &hidden_file->f_pos);
	set_fs(old_fs);

	kunmap(page);

	if(err == 0) {
		flush_dcache_page(page);
		SetPageUptodate(page);
	} else {
		ClearPageUptodate(page);
		err = 0;
	}
	
#else
	inode = file->f_dentry->d_inode;
	hidden_inode = itohi(inode);

	page->mapping = hidden_inode->i_mapping;
	err = hidden_inode->i_mapping->a_ops->readpage(hidden_file, page);
	page->mapping = inode->i_mapping;

#endif

	trace_readpage_post(stoptr(file->f_dentry->d_inode->i_sb), file, page, err, &context);

#ifdef TRACEFS_MMAP_CHANGES
	unlock_page(page);
	tracefs_read_unlock(file->f_dentry->d_sb);
#endif
	print_exit_status(err);
	return err;
}


STATIC int
tracefs_prepare_write(file_t *file, page_t *page, unsigned from, unsigned to)
{
	int err = 0;
	inode_t *inode;
	inode_t *hidden_inode;
	file_t* hidden_file = NULL;
	int page_lock_set = 0;
	void *context;

	print_entry_location();
	tracefs_read_lock(file->f_dentry->d_sb);
	trace_prepare_write_pre(stoptr(file->f_dentry->d_inode->i_sb), file, page, from, to, &context);
#ifndef TRACEFS_MMAP_CHANGES
	inode = file->f_dentry->d_inode;
	hidden_inode = itohi(inode);
	if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
	ASSERT(hidden_file != NULL);

	/*
	 * Tim: if the lower-page is not locked, lock it before calling the lower
	 * file-system's prepare_write
	 */
    if(!PageLocked(page)) {
	page_lock_set = 1;
	SetPageLocked(page);
    }

    page->mapping = hidden_inode->i_mapping;
    err = hidden_inode->i_mapping->a_ops->prepare_write(hidden_file, page, from, to);
    page->mapping = inode->i_mapping;

    if(page_lock_set)
		ClearPageLocked(page);

out:
#endif
    trace_prepare_write_post(stoptr(file->f_dentry->d_inode->i_sb), file, page, from, to, err, &context);
    print_exit_status(err);
	tracefs_read_unlock(file->f_dentry->d_sb);
    return err;
}



STATIC int
tracefs_commit_write(file_t *file, page_t *page, unsigned from, unsigned to)
{
    int err = -EINVAL;
    inode_t *inode;
    inode_t *hidden_inode;
    file_t *hidden_file = NULL;
    unsigned pos;
    void *context;
	unsigned bytes = to - from;
    char *page_data = NULL;
    mm_segment_t old_fs;

    print_entry_location();
	tracefs_read_lock(file->f_dentry->d_sb);
    trace_commit_write_pre(stoptr(file->f_dentry->d_inode->i_sb), file, page, from, to, &context);

    inode = page->mapping->host;
    hidden_inode = itohi(inode);

    ASSERT(file != NULL);

    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);

#ifdef TRACEFS_MMAP_CHANGES
	page_data = (char *)kmap(page);
    hidden_file->f_pos = tracefs_page_offset(page) + from;

    /*
     * SP: I use vfs_write instead of copying page data and the
     * prepare_write/commit_write combo because file system's like
     * GFS/OCFS2 don't like things touching those directly,
     * calling the underlying write op, while a little bit slower, will
     * call all the FS specific code as well
     */
    old_fs = get_fs();
    set_fs(KERNEL_DS);
    err = vfs_write(hidden_file, page_data + from, bytes,
            &hidden_file->f_pos);
    set_fs(old_fs);

    kunmap(page);
	if (err < 0)
        goto out;

#else
    page->mapping = hidden_inode->i_mapping;

    //XXX: Tim: Our caller is responsible for waiting
    //ASSERT(!PageWriteback(page));
    err = hidden_inode->i_mapping->a_ops->commit_write(hidden_file, page, from, to);
    page->mapping = inode->i_mapping;

#endif

    if (err >= 0) {
		inode->i_blocks = hidden_inode->i_blocks;
	/* we may have to update i_size */
	pos = tracefs_page_offset(page) + to;
	if (pos > inode->i_size)
		i_size_write(inode, pos);

	/*
	 * update mtime and ctime of lower level file system
	 * tracefs' mtime and ctime are updated by generic_file_write
	 */
	hidden_inode->i_mtime = hidden_inode->i_ctime = CURRENT_TIME;
	//mark_inode_dirty_sync(inode);
    }

out:
    trace_commit_write_post(stoptr(file->f_dentry->d_inode->i_sb), file, page, from, to, err, &context);
	tracefs_read_unlock(file->f_dentry->d_sb);
    print_exit_status(err);
    return err;			/* assume all is ok */
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
STATIC int
tracefs_bmap(struct address_space *mapping, long block)
#else
STATIC sector_t
tracefs_bmap(struct address_space *mapping,sector_t block)
#endif
{
    int err = 0;
    inode_t *inode;
    inode_t *hidden_inode;

    print_entry_location();

    inode = (inode_t *) mapping->host;
    hidden_inode = itohi(inode);

    if (hidden_inode->i_mapping->a_ops->bmap)
	err = hidden_inode->i_mapping->a_ops->bmap(hidden_inode->i_mapping, block);
    print_exit_location();
    return err;
}

/*
 * XXX: we may not need this function if not FIST_FILTER_DATA.
 * FIXME: for FIST_FILTER_SCA, get all lower pages and sync them each.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
STATIC int
#else
STATIC void
#endif
tracefs_sync_page(page_t *page)
{
    int err = 0;
    inode_t *inode;
    inode_t *hidden_inode;
    void* context;

    print_entry_location();

    inode = page->mapping->host;
    hidden_inode = itohi(inode);

    page->mapping = hidden_inode->i_mapping;
    if (page->mapping && page->mapping->a_ops && page->mapping->a_ops->sync_page) {

		trace_sync_page_pre(stoptr(inode->i_sb),page,&context);
		wait_on_page_writeback(page);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
		err =  page->mapping->a_ops->sync_page(page);
#else
		page->mapping->a_ops->sync_page(page);
		err = 0;
#endif

		trace_sync_page_post(stoptr(inode->i_sb),page,err,&context);
    }
    page->mapping = inode->i_mapping;

out:
    print_exit_status(err);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
    return err;
#else
    return;
#endif

}


struct address_space_operations tracefs_aops =
{
    writepage:		tracefs_writepage,
    readpage:		tracefs_readpage,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    //readpages:          tracefs_readpages,
#endif
    prepare_write:	tracefs_prepare_write,
    commit_write:	tracefs_commit_write,
    bmap:		tracefs_bmap,
    sync_page:		tracefs_sync_page,
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
