/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: super.c,v 1.25 2007/02/01 07:28:25 gopalas Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_tracefs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "tracefs.h"
#include "tracer.h"
#include "namecache.h"
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
STATIC void
tracefs_read_inode(inode_t *inode)
{
	static struct address_space_operations tracefs_empty_aops;

	print_entry_location();
	//trace_read_inode_pre(stoptr(inode->i_sb), inode);

	itopdlh(inode) = KMALLOC(sizeof(struct tracefs_inode_info), GFP_KERNEL);
	if (!itopd(inode)) {
		printk("<0>%s:%s:%d: No kernel memory!\n", __FILE__, __FUNCTION__, __LINE__);
		ASSERT(NULL);
	}
	itohi(inode) = NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	inode->i_version = ++event;	/* increment inode version */
#else
	inode->i_version++; /* increment inode version */
#endif
	inode->i_op = &tracefs_main_iops;
#if 0
	/*
	 * XXX: To export a file system via NFS, it has to have the
	 * FS_REQUIRES_DEV flag, so turn it on.  But should we inherit it from
	 * the lower file system, or can we allow our file system to be exported
	 * even if the lower one cannot be natively exported.
	 */
	inode->i_sb->s_type->fs_flags |= FS_REQUIRES_DEV;
	/*
	 * OK, the above was a hack, which is now turned off because it may
	 * cause a panic/oops on some systems.  The correct way to export a
	 * "nodev" filesystem is via using nfs-utils > 1.0 and the "fsid=" export
	 * parameter, which requires 2.4.20 or later.
	 */
#endif
	/* I don't think ->a_ops is ever allowed to be NULL */
	if (mmap_support(inode->i_sb)) {
		inode->i_mapping->a_ops = &tracefs_aops;
		fist_dprint(7, "setting inode %p to tracefs_aops (%p)\n", inode, tracefs_aops);
		inode->i_fop = &tracefs_mmap_fops;
	} else {
		inode->i_fop = &tracefs_main_fops;
		inode->i_mapping->a_ops = &tracefs_empty_aops;
		fist_dprint(7, "setting inode %p a_ops to empty (%p)\n",
			inode, inode->i_mapping->a_ops);
	}

	print_exit_location();
}


#if defined(FIST_DEBUG) || defined(FIST_FILTER_SCA)
/*
 * No need to call write_inode() on the lower inode, as it
 * will have been marked 'dirty' anyway. But we might need
 * to write some of our own stuff to disk.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
STATIC void
#else
STATIC int
#endif
tracefs_write_inode(inode_t *inode, int sync)
{
    void *context;
    print_entry_location();
    trace_write_inode_pre(stoptr(inode->i_sb), inode, sync, &context);
    trace_write_inode_post(stoptr(inode->i_sb), inode, sync, &context);
    print_exit_location();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    return 0;
#endif
}
#endif /* defined(FIST_DEBUG) || defined(FIST_FILTER_SCA) */


STATIC void
tracefs_put_inode(inode_t *inode)
{
    void *context;
    print_entry_location();
    trace_put_inode_pre(stoptr(inode->i_sb), inode, &context);
    fist_dprint(8, "%s i_count = %d, i_nlink = %d\n", __FUNCTION__,
		atomic_read(&inode->i_count), inode->i_nlink);
    /*
     * This is really funky stuff:
     * Basically, if i_count == 1, iput will then decrement it and this inode will be destroyed.
     * It is currently holding a reference to the hidden inode.
     * Therefore, it needs to release that reference by calling iput on the hidden inode.
     * iput() _will_ do it for us (by calling our clear_inode), but _only_ if i_nlink == 0.
     * The problem is, NFS keeps i_nlink == 1 for silly_rename'd files.
     * So we must for our i_nlink to 0 here to trick iput() into calling our clear_inode.
     */
    if (atomic_read(&inode->i_count) == 1)
	inode->i_nlink = 0;
    trace_put_inode_post(stoptr(inode->i_sb), inode, &context);
    print_exit_location();
}


/*
 * we now define delete_inode, because there are two VFS paths that may
 * destroy an inode: one of them calls clear inode before doing everything
 * else that's needed, and the other is fine.  This way we truncate the inode
 * size (and its pages) and then clear our own inode, which will do an iput
 * on our and the lower inode.
 */
STATIC void
tracefs_delete_inode(inode_t *inode)
{
    void *context;
    print_entry_location();
    trace_delete_inode_pre(stoptr(inode->i_sb), inode, &context);

    fist_checkinode(inode, "tracefs_delete_inode IN");
    truncate_inode_pages(&inode->i_data, 0);
    inode->i_size = 0;		/* every f/s seems to do that */
    clear_inode(inode);

    trace_delete_inode_post(stoptr(inode->i_sb), inode, &context);
    print_exit_location();
}


/* final actions when unmounting a file system */
STATIC void
tracefs_put_super(super_block_t *sb)
{
    print_entry_location();

    if (stopd(sb)) {
	mntput(stopd(sb)->hidden_mnt);
	if (stopnc(sb)) {
	    namecache_release(stopnc(sb));
	}
	KFREE(stopd(sb));
	stopdlh(sb) = NULL;
    }
    fist_dprint(6, "tracefs: released super\n");

    print_exit_location();
}


#ifdef NOT_NEEDED
/*
 * This is called in do_umount before put_super.
 * The superblock lock is not held yet.
 * We probably do not need to define this or call write_super
 * on the hidden_sb, because sync_supers() will get to hidden_sb
 * sooner or later.  But it is also called from file_fsync()...
 */
STATIC void
tracefs_write_super(super_block_t *sb)
{
    return;
}
#endif /* NOT_NEEDED */


STATIC int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
tracefs_statfs(super_block_t *sb, struct statfs *buf)
#else
tracefs_statfs(super_block_t *sb, struct kstatfs *buf)
#endif
{
    int err = 0;
    void *context;
    super_block_t *hidden_sb;

    print_entry_location();
    trace_statfs_pre(stoptr(sb), buf, &context);
    hidden_sb = stohs(sb);
    err = vfs_statfs(hidden_sb, buf);

    trace_statfs_post(stoptr(sb), buf, err, &context);
    print_exit_status(err);
    return err;
}


/*
 * XXX: not implemented.  This is not allowed yet.
 * Should we call this on the hidden_sb?  Probably not.
 */
STATIC int
tracefs_remount_fs(super_block_t *sb, int *flags, char *data)
{
    return -ENOSYS;
}


/*
 * Called by iput() when the inode reference count reached zero
 * and the inode is not hashed anywhere.  Used to clear anything
 * that needs to be, before the inode is completely destroyed and put
 * on the inode free list.
 */
STATIC void
tracefs_clear_inode(inode_t *inode)
{

    void *context;
    print_entry_location();
    trace_clear_inode_pre(stoptr(inode->i_sb), inode, &context);
    fist_checkinode(inode, "tracefs_clear_inode IN");
    /*
     * Remove the inode from the trace pattern name cache
     * XXX: May have to clear the name cache using a different
     * strategy since this is a potentially expensive operation
     */
    delete_inode_global(stopnc(inode->i_sb), inode);


    /*
     * Decrement a reference to a hidden_inode, which was incremented
     * by our read_inode when it was created initially.
     */
    iput(itohi(inode));
    // XXX: why this assertion fails?
    // because it doesn't like us
    // ASSERT((inode->i_state & I_DIRTY) == 0);
    KFREE(itopd(inode));
    itopdlh(inode) = NULL;

    trace_clear_inode_post(stoptr(inode->i_sb), inode, &context);
    print_exit_location();
}


/*
 * Called in do_umount() if the MNT_FORCE flag was used and this
 * function is defined.  See comment in linux/fs/super.c:do_umount().
 * Used only in nfs, to kill any pending RPC tasks, so that subsequent
 * code can actually succeed and won't leave tasks that need handling.
 *
 * PS. I wonder if this is somehow useful to undo damage that was
 * left in the kernel after a user level file server (such as amd)
 * dies.
 */
STATIC void
tracefs_umount_begin(super_block_t *sb)
{
    super_block_t *hidden_sb;

    print_entry_location();

    hidden_sb = stohs(sb);

    if (hidden_sb->s_op->umount_begin)
	hidden_sb->s_op->umount_begin(hidden_sb);

    print_exit_location();
}

#undef clear_inode
struct super_operations tracefs_sops =
{
    read_inode:		tracefs_read_inode,
#if defined(FIST_DEBUG) || defined(FIST_FILTER_SCA)
    write_inode:	tracefs_write_inode,
#endif /* defined(FIST_DEBUG) || defined(FIST_FILTER_SCA) */
    put_inode:		tracefs_put_inode,
    delete_inode:	tracefs_delete_inode,
    put_super:		tracefs_put_super,
    statfs:		tracefs_statfs,
    remount_fs:		tracefs_remount_fs,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    clear_inode:	tracefs_clear_inode,
#endif
    umount_begin:	tracefs_umount_begin,
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
